import pyxel

TRANSPARENT_COLOR = 2
SCROLL_BORDER_X = 80
TILE_FLOOR = (1, 0)
TILE_SPAWN1 = (0, 1)
TILE_SPAWN2 = (1, 1)
TILE_SPAWN3 = (2, 1)
WALL_TILE_X = 4

scroll_x = 0
player = None
enemies = []


def get_tile(tile_x, tile_y):
    return pyxel.tilemap(0).pget(tile_x, tile_y)


def detect_collision(x, y, dy):
    x1 = x // 8
    y1 = y // 8
    x2 = (x + 8 - 1) // 8
    y2 = (y + 8 - 1) // 8
    for yi in range(y1, y2 + 1):
        for xi in range(x1, x2 + 1):
            if get_tile(xi, yi)[0] >= WALL_TILE_X:
                return True
    if dy > 0 and y % 8 == 1:
        for xi in range(x1, x2 + 1):
            if get_tile(xi, y1 + 1) == TILE_FLOOR:
                return True
    return False


def push_back(x, y, dx, dy):
    abs_dx = abs(dx)
    abs_dy = abs(dy)
    if abs_dx > abs_dy:
        sign = 1 if dx > 0 else -1
        for _ in range(abs_dx):
            if detect_collision(x + sign, y, dy):
                break
            x += sign
        sign = 1 if dy > 0 else -1
        for _ in range(abs_dy):
            if detect_collision(x, y + sign, dy):
                break
            y += sign
    else:
        sign = 1 if dy > 0 else -1
        for _ in range(abs_dy):
            if detect_collision(x, y + sign, dy):
                break
            y += sign
        sign = 1 if dx > 0 else -1
        for _ in range(abs_dx):
            if detect_collision(x + sign, y, dy):
                break
            x += sign
    return x, y, dx, dy


def is_wall(x, y):
    tile = get_tile(x // 8, y // 8)
    return tile == TILE_FLOOR or tile[0] >= WALL_TILE_X


def spawn_enemy(left_x, right_x):
    left_x = pyxel.ceil(left_x / 8)
    right_x = pyxel.floor(right_x / 8)
    for x in range(left_x, right_x + 1):
        for y in range(16):
            tile = get_tile(x, y)
            if tile == TILE_SPAWN1:
                enemies.append(Enemy1(x * 8, y * 8))
            elif tile == TILE_SPAWN2:
                enemies.append(Enemy2(x * 8, y * 8))
            elif tile == TILE_SPAWN3:
                enemies.append(Enemy3(x * 8, y * 8))


def cleanup_list(list):
    i = 0
    while i < len(list):
        elem = list[i]
        if elem.is_alive:
            i += 1
        else:
            list.pop(i)


class Player:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.dx = 0
        self.dy = 0
        self.direction = 1
        self.is_falling = False

    def update(self):
        global scroll_x
        last_y = self.y
        if pyxel.btn(pyxel.KEY_LEFT):
            self.dx = -2
            self.direction = -1
        if pyxel.btn(pyxel.KEY_RIGHT):
            self.dx = 2
            self.direction = 1
        self.dy = min(self.dy + 1, 3)
        if pyxel.btnp(pyxel.KEY_SPACE):
            self.dy = -6
            pyxel.play(3, 8)
        self.x, self.y, self.dx, self.dy = push_back(self.x, self.y, self.dx, self.dy)
        if self.x < scroll_x:
            self.x = scroll_x
        if self.y < 0:
            self.y = 0
        self.dx = int(self.dx * 0.8)
        self.is_falling = self.y > last_y

        if self.x > scroll_x + SCROLL_BORDER_X:
            last_scroll_x = scroll_x
            scroll_x = min(self.x - SCROLL_BORDER_X, 240 * 8)
            spawn_enemy(last_scroll_x + 128, scroll_x + 127)
        if self.y >= pyxel.height:
            game_over()

    def draw(self):
        u = (2 if self.is_falling else pyxel.frame_count // 3 % 2) * 8
        w = 8 if self.direction > 0 else -8
        pyxel.blt(self.x, self.y, 0, u, 16, w, 8, TRANSPARENT_COLOR)


class Enemy1:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.dx = 0
        self.dy = 0
        self.direction = -1
        self.is_alive = True

    def update(self):
        self.dx = self.direction
        self.dy = min(self.dy + 1, 3)
        if self.direction < 0 and is_wall(self.x - 1, self.y + 4):
            self.direction = 1
        elif self.direction > 0 and is_wall(self.x + 8, self.y + 4):
            self.direction = -1
        self.x, self.y, self.dx, self.dy = push_back(self.x, self.y, self.dx, self.dy)

    def draw(self):
        u = pyxel.frame_count // 4 % 2 * 8
        w = 8 if self.direction > 0 else -8
        pyxel.blt(self.x, self.y, 0, u, 24, w, 8, TRANSPARENT_COLOR)


class Enemy2:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.dx = 0
        self.dy = 0
        self.direction = 1
        self.is_alive = True

    def update(self):
        self.dx = self.direction
        self.dy = min(self.dy + 1, 3)
        if is_wall(self.x, self.y + 8) or is_wall(self.x + 7, self.y + 8):
            if self.direction < 0 and (
                is_wall(self.x - 1, self.y + 4) or not is_wall(self.x - 1, self.y + 8)
            ):
                self.direction = 1
            elif self.direction > 0 and (
                is_wall(self.x + 8, self.y + 4) or not is_wall(self.x + 7, self.y + 8)
            ):
                self.direction = -1
        self.x, self.y, self.dx, self.dy = push_back(self.x, self.y, self.dx, self.dy)

    def draw(self):
        u = pyxel.frame_count // 4 % 2 * 8 + 16
        w = 8 if self.direction > 0 else -8
        pyxel.blt(self.x, self.y, 0, u, 24, w, 8, TRANSPARENT_COLOR)


class Enemy3:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.time_to_fire = 0
        self.is_alive = True

    def update(self):
        self.time_to_fire -= 1
        if self.time_to_fire <= 0:
            dx = player.x - self.x
            dy = player.y - self.y
            sq_dist = dx * dx + dy * dy
            if sq_dist < 60**2:
                dist = pyxel.sqrt(sq_dist)
                enemies.append(Enemy3Bullet(self.x, self.y, dx / dist, dy / dist))
                self.time_to_fire = 60

    def draw(self):
        u = pyxel.frame_count // 8 % 2 * 8
        pyxel.blt(self.x, self.y, 0, u, 32, 8, 8, TRANSPARENT_COLOR)


class Enemy3Bullet:
    def __init__(self, x, y, dx, dy):
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.is_alive = True

    def update(self):
        self.x += self.dx
        self.y += self.dy

    def draw(self):
        u = pyxel.frame_count // 2 % 2 * 8 + 16
        pyxel.blt(self.x, self.y, 0, u, 32, 8, 8, TRANSPARENT_COLOR)


class App:
    def __init__(self):
        pyxel.init(128, 128, title="Pyxel Platformer")
        pyxel.load("assets/platformer.pyxres")

        # Change enemy spawn tiles invisible
        pyxel.image(0).rect(0, 8, 24, 8, TRANSPARENT_COLOR)

        global player
        player = Player(0, 0)
        spawn_enemy(0, 127)
        pyxel.playm(0, loop=True)
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btn(pyxel.KEY_Q):
            pyxel.quit()

        player.update()
        for enemy in enemies:
            if abs(player.x - enemy.x) < 6 and abs(player.y - enemy.y) < 6:
                game_over()
                return
            enemy.update()
            if enemy.x < scroll_x - 8 or enemy.x > scroll_x + 160 or enemy.y > 160:
                enemy.is_alive = False
        cleanup_list(enemies)

    def draw(self):
        pyxel.cls(0)

        # Draw level
        pyxel.camera()
        pyxel.bltm(0, 0, 0, (scroll_x // 4) % 128, 128, 128, 128)
        pyxel.bltm(0, 0, 0, scroll_x, 0, 128, 128, TRANSPARENT_COLOR)

        # Draw characters
        pyxel.camera(scroll_x, 0)
        player.draw()
        for enemy in enemies:
            enemy.draw()


def game_over():
    global scroll_x, enemies
    scroll_x = 0
    player.x = 0
    player.y = 0
    player.dx = 0
    player.dy = 0
    enemies = []
    spawn_enemy(0, 127)
    pyxel.play(3, 9)


App()



import pyxel

from .editor_base import EditorBase
from .field_cursor import FieldCursor
from .music_field import MusicField
from .settings import EDITOR_IMAGE, MAX_MUSIC_LENGTH, TEXT_LABEL_COLOR
from .sound_selector import SoundSelector
from .widgets import ImageButton, ImageToggleButton, NumberPicker


class MusicEditor(EditorBase):
    """
    Variables:
        music_no_var
        should_loop_var
        is_playing_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.copy_var("help_message_var", parent)

        # Initialize is_playing_var
        self.new_var("is_playing_var", False)

        # Initialize field cursor
        self.field_cursor = FieldCursor(
            max_field_length=MAX_MUSIC_LENGTH,
            field_wrap_length=16,
            get_field=self.get_field,
            add_pre_history=self.add_pre_history,
            add_post_history=self.add_post_history,
            cross_filed_copying=True,
        )

        # Initialize music picker
        self._music_picker = NumberPicker(
            self, 45, 17, min_value=0, max_value=pyxel.NUM_MUSICS - 1, value=0
        )
        self.add_number_picker_help(self._music_picker)
        self.copy_var("music_no_var", self._music_picker, "value_var")

        # Initialize play button
        self._play_button = ImageButton(
            self,
            185,
            17,
            img=EDITOR_IMAGE,
            u=126,
            v=0,
        )
        self._play_button.add_event_listener("press", self.__on_play_button_press)
        self._play_button.add_event_listener(
            "mouse_hover", self.__on_play_button_mouse_hover
        )

        # Initialize stop button
        self._stop_button = ImageButton(
            self, 195, 17, img=EDITOR_IMAGE, u=135, v=0, is_enabled=False
        )
        self._stop_button.add_event_listener("press", self.__on_stop_button_press)
        self._stop_button.add_event_listener(
            "mouse_hover", self.__on_stop_button_mouse_hover
        )

        # Initialize loop button
        self._loop_button = ImageToggleButton(
            self, 205, 17, img=EDITOR_IMAGE, u=144, v=0, is_checked=False
        )
        self._loop_button.add_event_listener(
            "mouse_hover", self.__on_loop_button_mouse_hover
        )
        self.copy_var("should_loop_var", self._loop_button, "is_checked_var")

        # Initialize music field
        self._music_field = [MusicField(self, 11, 29 + i * 25, i) for i in range(4)]

        # Initialize sound selector
        self._sound_selector = SoundSelector(self)

        # Set event listeners
        self.add_event_listener("undo", self.__on_undo)
        self.add_event_listener("redo", self.__on_redo)
        self.add_event_listener("hide", self.__on_hide)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def play_pos(self, ch):
        return self._play_pos[ch]

    def get_field(self, index):
        if index >= pyxel.NUM_CHANNELS:
            return
        music = pyxel.music(self.music_no_var)
        return music.snds_list[index]

    def add_pre_history(self, x, y):
        self._history_data = data = {}
        data["music_no"] = self.music_no_var
        data["old_cursor_pos"] = (x, y)
        data["old_field"] = self.field_cursor.field.to_list()

    def add_post_history(self, x, y):
        data = self._history_data
        data["new_cursor_pos"] = (x, y)
        data["new_field"] = self.field_cursor.field.to_list()
        if data["old_field"] != data["new_field"]:
            self.add_history(self._history_data)

    def _play(self, is_partial):
        self.is_playing_var = True
        self._music_picker.is_enabled_var = False
        self._play_button.is_enabled_var = False
        self._stop_button.is_enabled_var = True
        self._loop_button.is_enabled_var = False
        tick = 0
        if is_partial:
            for i in range(self.field_cursor.x):
                music = pyxel.music(self.music_no_var)
                sound = pyxel.sound(music.snds_list[self.field_cursor.y][i])
                tick += len(sound.notes) * sound.speed
        pyxel.playm(self.music_no_var, tick=tick, loop=self.should_loop_var)

    def _stop(self):
        self.is_playing_var = False
        self._music_picker.is_enabled_var = True
        self._play_button.is_enabled_var = True
        self._stop_button.is_enabled_var = False
        self._loop_button.is_enabled_var = True
        pyxel.stop()

    def __on_undo(self, data):
        self._stop()
        self.music_no_var = data["music_no"]
        self.field_cursor.move_to(*data["old_cursor_pos"], False)
        self.field_cursor.field.from_list(data["old_field"])

    def __on_redo(self, data):
        self._stop()
        self.music_no_var = data["music_no"]
        self.field_cursor.move_to(*data["new_cursor_pos"], False)
        self.field_cursor.field.from_list(data["new_field"])

    def __on_hide(self):
        self._stop()

    def __on_update(self):
        if self.is_playing_var:
            self.is_playing_var = None
            for i in range(pyxel.NUM_CHANNELS):
                if pyxel.play_pos(i) is not None:
                    self.is_playing_var = True
                    break
        if pyxel.btnp(pyxel.KEY_SPACE):
            if self.is_playing_var:
                self._stop_button.is_pressed_var = True
            else:
                self._play_button.is_pressed_var = True
        if self.is_playing_var:
            return
        if not self._play_button.is_enabled_var:
            self._stop()
        if self._loop_button.is_enabled_var and pyxel.btnp(pyxel.KEY_L):
            self.should_loop_var = not self.should_loop_var
        self.field_cursor.process_input()

    def __on_draw(self):
        self.draw_panel(11, 16, 218, 9)
        pyxel.text(23, 18, "MUSIC", TEXT_LABEL_COLOR)

    def __on_play_button_press(self):
        self._play(pyxel.btn(pyxel.KEY_SHIFT))

    def __on_stop_button_press(self):
        self._stop()

    def __on_play_button_mouse_hover(self, x, y):
        self.help_message_var = "PLAY:SPACE PART-PLAY:SHIFT+SPACE"

    def __on_stop_button_mouse_hover(self, x, y):
        self.help_message_var = "STOP:SPACE"

    def __on_loop_button_mouse_hover(self, x, y):
        self.help_message_var = "LOOP:L"



import pyxel

from .settings import (
    EDITOR_IMAGE,
    MAX_SOUND_LENGTH,
    PIANO_ROLL_BACKGROUND_COLOR,
    PIANO_ROLL_CURSOR_EDIT_COLOR,
    PIANO_ROLL_CURSOR_PLAY_COLOR,
    PIANO_ROLL_CURSOR_SELECT_COLOR,
    PIANO_ROLL_NOTE_COLOR,
    PIANO_ROLL_REST_COLOR,
)
from .widgets import Widget
from .widgets.settings import WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME


class PianoRoll(Widget):
    """
    Variables:
        note_var
        is_playing_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent, 30, 25, 193, 123)
        self._press_x = 0
        self._press_y = 0
        self.field_cursor = parent.field_cursor
        self.get_field = parent.get_field
        self.add_pre_history = parent.add_pre_history
        self.add_post_history = parent.add_post_history
        self.get_field_help_message = parent.get_field_help_message
        self.copy_var("note_var", parent)
        self.copy_var("is_playing_var", parent)
        self.copy_var("help_message_var", parent)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_click", self.__on_mouse_click)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_view(self, x, y):
        x = min(max((x - self.x - 1) // 4, 0), MAX_SOUND_LENGTH - 1)
        y = min(max(59 - (y - self.y - 1) // 2, -1), 59)
        return x, y

    def _set_note(self, x, y):
        self.add_pre_history(x, 0)
        self.field_cursor.move_to(x, 0, pyxel.btn(pyxel.KEY_SHIFT))
        field = self.field_cursor.field
        field_len = len(field)
        if x < field_len:
            field[x] = y
        else:
            lst = field.to_list()
            lst.extend([-1] * (x - field_len) + [y])
            field.from_list(lst)
        self.add_post_history(x, 0)

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        x, y = self._screen_to_view(x, y)
        self._press_x = x
        self._press_y = y
        self.field_cursor.move_to(x, 0, pyxel.btn(pyxel.KEY_SHIFT))

    def __on_mouse_drag(self, key, x, y, dx, dy):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        x, y = self._screen_to_view(x, y)
        if x > self._press_x:
            step = 1
        elif x < self._press_x:
            step = -1
        else:
            if y != self._press_y:
                self._set_note(x, y)
                self._press_x = x
                self._press_y = y
            return
        dx = x - self._press_x
        dy = y - self._press_y
        alpha = dy / dx
        for i in range(0, dx + step, step):
            self._set_note(self._press_x + i, round(self._press_y + alpha * i))
        self._press_x = x
        self._press_y = y

    def __on_mouse_click(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        x, y = self._screen_to_view(x, y)
        self._set_note(x, y)

    def __on_mouse_hover(self, x, y):
        self.help_message_var = self.get_field_help_message()

    def __on_update(self):
        if self.field_cursor.y > 0 or self.is_playing_var:
            return
        if (
            pyxel.btnp(pyxel.KEY_RETURN, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME)
            or pyxel.btnp(pyxel.KEY_KP_ENTER, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME)
        ) and self.note_var is not None:
            self.field_cursor.insert(self.note_var)

    def __on_draw(self):
        # Draw frame
        pyxel.rect(self.x, self.y, self.width, self.height, 7)
        if self.is_playing_var:
            x = pyxel.play_pos(0)[1] * 4 + 31
            pyxel.rect(x, 25, 3, 123, PIANO_ROLL_CURSOR_PLAY_COLOR)
        elif self.field_cursor.y == 0:
            x = self.field_cursor.x * 4 + 31
            w = self.field_cursor.width * 4 - 1
            col = (
                PIANO_ROLL_CURSOR_SELECT_COLOR
                if self.field_cursor.is_selecting
                else PIANO_ROLL_CURSOR_EDIT_COLOR
            )
            pyxel.rect(x, 25, w, 123, col)
        pyxel.blt(
            self.x,
            self.y,
            EDITOR_IMAGE,
            0,
            7,
            193,
            72,
            PIANO_ROLL_BACKGROUND_COLOR,
        )
        pyxel.blt(
            self.x,
            self.y + 72,
            EDITOR_IMAGE,
            0,
            7,
            193,
            51,
            PIANO_ROLL_BACKGROUND_COLOR,
        )

        # Draw notes
        notes = self.get_field(0)
        for i, note in enumerate(notes):
            pyxel.rect(
                i * 4 + 31,
                143 - note * 2,
                3,
                3,
                PIANO_ROLL_NOTE_COLOR if note >= 0 else PIANO_ROLL_REST_COLOR,
            )



import pyxel

from .settings import (
    EDITOR_IMAGE,
    MUSIC_FIELD_BACKGROUND_COLOR,
    MUSIC_FIELD_CURSOR_EDIT_COLOR,
    MUSIC_FIELD_CURSOR_PLAY_COLOR,
    MUSIC_FIELD_CURSOR_SELECT_COLOR,
    MUSIC_FIELD_SOUND_NORMAL_COLOR,
    MUSIC_FIELD_SOUND_SELECT_COLOR,
    TEXT_LABEL_COLOR,
)
from .widgets import Widget


class MusicField(Widget):
    """
    Variables:
        is_playing_var
        help_message_var
    """

    def __init__(self, parent, x, y, ch):
        super().__init__(parent, x, y, 218, 21)
        self._ch = ch
        self.field_cursor = parent.field_cursor
        self.get_field = parent.get_field
        self.copy_var("is_playing_var", parent)
        self.copy_var("help_message_var", parent)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("draw", self.__on_draw)

    @property
    def data(self):
        return self.get_field(self._ch)

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        x -= self.x + 21
        y -= self.y + 2
        if x < 0 or y < 0 or x > 188 or y > 16 or x % 12 > 8 or y % 10 > 6:
            return
        self.field_cursor.move_to(
            x // 12 + (y // 10) * 16, self._ch, pyxel.btn(pyxel.KEY_SHIFT)
        )

    def __on_mouse_hover(self, x, y):
        self.help_message_var = "SOUND:SOUND_BUTTON/BS/DEL"

    def __on_draw(self):
        # Draw frame
        self.draw_panel(self.x, self.y, self.width, self.height)
        pyxel.text(self.x + 5, self.y + 8, f"CH{self._ch}", TEXT_LABEL_COLOR)
        pyxel.blt(
            self.x + 20,
            self.y + 1,
            EDITOR_IMAGE,
            0,
            102,
            191,
            19,
            MUSIC_FIELD_BACKGROUND_COLOR,
        )

        # Draw cursor
        if self.is_playing_var:
            play_pos = pyxel.play_pos(self._ch)
            if play_pos is None:
                cursor_x = -1
                cursor_y = -1
            else:
                cursor_x = play_pos[0]
                cursor_y = self._ch
                cursor_width = 1
                cursor_col = MUSIC_FIELD_CURSOR_PLAY_COLOR
        else:
            cursor_x = self.field_cursor.x
            cursor_y = self.field_cursor.y
            cursor_width = self.field_cursor.width
            cursor_col = (
                MUSIC_FIELD_CURSOR_SELECT_COLOR
                if self.field_cursor.is_selecting
                else MUSIC_FIELD_CURSOR_EDIT_COLOR
            )
        if cursor_y == self._ch:
            for i in range(len(self.data) + 1):
                if cursor_x <= i < cursor_x + cursor_width:
                    x = self.x + (i % 16) * 12 + 21
                    y = self.y + (cursor_y - self._ch + i // 16) * 10 + 2
                    pyxel.rect(x, y, 9, 7, cursor_col)

        # Draw sounds
        for i in range(len(self.data)):
            x = self.x + 22 + (i % 16) * 12
            y = self.y + (i // 16) * 10 + 3
            col = (
                MUSIC_FIELD_SOUND_SELECT_COLOR
                if cursor_y == self._ch and cursor_x <= i < cursor_x + cursor_width
                else MUSIC_FIELD_SOUND_NORMAL_COLOR
            )
            pyxel.text(x, y, f"{self.data[i]:0>2}", col)



import pyxel

from .widget import Widget


class ColorPicker(Widget):
    """
    Variables:
        value_var

    Events:
        change (value)
    """

    def __init__(self, parent, x, y, value, *, with_shadow=False, **kwargs):
        super().__init__(parent, x, y, 65, 17, **kwargs)
        self._with_shadow = with_shadow

        # Initialize value_var
        self.new_var("value_var", value)
        self.add_var_event_listener("value_var", "change", self.__on_value_change)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("draw", self.__on_draw)

    def check_value(self, x, y):
        x -= self.x + 1
        y -= self.y + 1
        if (
            0 <= x <= self.width - 2
            and 0 <= y <= self.height - 2
            and x % 8 != 7
            and y != 7
        ):
            return (y // 8) * 8 + x // 8
        else:
            return None

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        value = self.check_value(x, y)
        if value is not None:
            self.value_var = value

    def __on_mouse_drag(self, key, x, y, dx, dy):
        self.__on_mouse_down(key, x, y)

    def __on_draw(self):
        self.draw_panel(
            self.x, self.y, self.width, self.height, with_shadow=self._with_shadow
        )

        # Draw colors
        for yi in range(2):
            for xi in range(8):
                pyxel.rect(self.x + xi * 8 + 1, self.y + yi * 8 + 1, 7, 7, yi * 8 + xi)

        # Draw cursor
        col = self.value_var
        pyxel.text(
            self.x + (col % 8) * 8 + 3,
            self.y + (col // 8) * 8 + 2,
            "+",
            7 if col < 6 else 0,
        )

    def __on_value_change(self, value):
        self.trigger_event("change", value)



import pyxel

from .settings import OCTAVE_BAR_BACKGROUND_COLOR, OCTAVE_BAR_COLOR
from .widgets import Widget


class OctaveBar(Widget):
    """
    Variables:
        octave_var
        help_message_var
    """

    def __init__(self, parent, x, y):
        super().__init__(parent, x, y, 4, 123)
        self.field_cursor = parent.field_cursor
        self.copy_var("octave_var", parent)
        self.copy_var("help_message_var", parent)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("draw", self.__on_draw)

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        if self.field_cursor.y > 0:
            self.field_cursor.move_to(self.field_cursor.x, 0, False)
        self.octave_var = min(max(3 - ((y - self.y - 12) // 24), 0), 3)

    def __on_mouse_drag(self, key, x, y, dx, dy):
        self.__on_mouse_down(key, x, y)

    def __on_mouse_hover(self, x, y):
        self.help_message_var = "OCTAVE:PAGEUP/PAGEDOWN"

    def __on_draw(self):
        x = self.x + 1
        y = self.y + 1 + (3 - self.octave_var) * 24
        pyxel.rect(self.x, self.y, self.width, self.height, OCTAVE_BAR_BACKGROUND_COLOR)
        pyxel.rect(x, y, 2, 47, OCTAVE_BAR_COLOR)



import pyxel


class App:
    def __init__(self):
        pyxel.init(64, 64, title="Perlin Noise", capture_scale=4)
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btnp(pyxel.KEY_Q):
            pyxel.quit()

    def draw(self):
        pyxel.cls(0)
        for y in range(64):
            for x in range(64):
                n = pyxel.noise(
                    x / 10,
                    y / 10,
                    pyxel.frame_count / 40,
                )
                if n > 0.4:
                    col = 7
                elif n > 0:
                    col = 6
                elif n > -0.4:
                    col = 12
                else:
                    col = 0
                pyxel.pset(x, y, col)


App()



import pyxel

from .canvas_panel import CanvasPanel
from .editor_base import EditorBase
from .image_viewer import ImageViewer
from .settings import EDITOR_IMAGE, TEXT_LABEL_COLOR, TOOL_PENCIL
from .widgets import ColorPicker, NumberPicker, RadioButton


class ImageEditor(EditorBase):
    """
    Variables:
        color_var
        tool_var
        image_no_var
        canvas_var
        focus_x_var
        focus_y_var
        help_message_var

    Events:
        undo (data)
        redo (data)
        drop (filename)
    """

    _COLOR_BUTTONS = (
        pyxel.KEY_1,
        pyxel.KEY_2,
        pyxel.KEY_3,
        pyxel.KEY_4,
        pyxel.KEY_5,
        pyxel.KEY_6,
        pyxel.KEY_7,
        pyxel.KEY_8,
    )

    def __init__(self, parent):
        super().__init__(parent)
        self.copy_var("help_message_var", parent)

        # Initialize canvas_var
        self.new_var("canvas_var", None)
        self.add_var_event_listener("canvas_var", "get", self.__on_canvas_get)

        # Initialize color picker
        self._color_picker = ColorPicker(self, 11, 156, 7, with_shadow=False)
        self._color_picker.add_event_listener(
            "mouse_hover", self.__on_color_picker_mouse_hover
        )
        self.copy_var("color_var", self._color_picker, "value_var")

        # Initialize tool button
        self._tool_button = RadioButton(
            self,
            81,
            161,
            img=EDITOR_IMAGE,
            u=63,
            v=0,
            num_buttons=7,
            value=TOOL_PENCIL,
        )
        self.add_tool_button_help(self._tool_button)
        self.copy_var("tool_var", self._tool_button, "value_var")

        # Initialize image picker
        self._image_picker = NumberPicker(
            self, 192, 161, min_value=0, max_value=pyxel.NUM_IMAGES - 1, value=0
        )
        self.add_number_picker_help(self._image_picker)
        self.copy_var("image_no_var", self._image_picker, "value_var")

        # Initialize image viewer
        self._image_viewer = ImageViewer(self)
        self.copy_var("focus_x_var", self._image_viewer)
        self.copy_var("focus_y_var", self._image_viewer)

        # Initialize canvas panel
        self._canvas_panel = CanvasPanel(self)

        # Set event listeners
        self.add_event_listener("undo", self.__on_undo)
        self.add_event_listener("redo", self.__on_redo)
        self.add_event_listener("drop", self.__on_drop)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def __on_canvas_get(self, value):
        return pyxel.image(self.image_no_var)

    def __on_color_picker_mouse_hover(self, x, y):
        self.help_message_var = "COLOR:1-8/SHIFT+1-8"

    def __on_undo(self, data):
        self.image_no_var = data["image_no"]
        self.focus_x_var, self.focus_y_var = data["focus_pos"]
        self.canvas_var.set_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, data["old_canvas"]
        )

    def __on_redo(self, data):
        self.image_no_var = data["image_no"]
        self.focus_x_var, self.focus_y_var = data["focus_pos"]
        self.canvas_var.set_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, data["new_canvas"]
        )

    def __on_drop(self, filename):
        pyxel.image(self.image_no_var).load(0, 0, filename)

    def __on_update(self):
        self.check_tool_button_shortcuts()

        # Check color shortcuts
        if not pyxel.btn(pyxel.KEY_ALT):
            for btn in self._COLOR_BUTTONS:
                if pyxel.btnp(btn):
                    col = btn - pyxel.KEY_1
                    if pyxel.btn(pyxel.KEY_SHIFT):
                        col += 8
                    self.color_var = col
                    break

    def __on_draw(self):
        self.draw_panel(11, 156, 136, 17)
        self.draw_panel(157, 156, 72, 17)
        pyxel.text(170, 162, "IMAGE", TEXT_LABEL_COLOR)



import pyxel

from .button import Button
from .settings import BUTTON_TEXT_COLOR


class TextButton(Button):
    """
    Variables:
        is_pressed_var

    Events:
        press
    """

    def __init__(self, parent, x, y, *, text, **kwargs):
        super().__init__(
            parent,
            x,
            y,
            len(text) * pyxel.FONT_WIDTH + 3,
            pyxel.FONT_HEIGHT + 1,
            **kwargs
        )
        self._text = text

        # Set event listeners
        self.add_event_listener("draw", self.__on_draw)

    def __on_draw(self):
        x = self.x
        y = self.y
        w = self.width
        h = self.height
        col = self.button_color
        pyxel.line(x + 1, y, x + w - 2, y, col)
        pyxel.rect(x, y + 1, w, h - 2, col)
        pyxel.line(x + 1, y + h - 1, x + w - 2, y + h - 1, col)
        pyxel.text(x + 2, y + 1, self._text, BUTTON_TEXT_COLOR)



import pyxel

BORDER_OFFSETS = [(-1, 0), (1, 0), (0, -1), (0, 1)]
BG_COLOR = 5


def pset_with_border(image, x, y, col, bcol):
    for x_offset, y_offset in BORDER_OFFSETS:
        image.pset(x + x_offset, y + y_offset, bcol)
    image.pset(x, y, col)


def line_with_border(image, x1, y1, x2, y2, col, bcol):
    for x_offset, y_offset in BORDER_OFFSETS:
        image.line(
            x1 + x_offset,
            y1 + y_offset,
            x2 + x_offset,
            y2 + y_offset,
            bcol,
        )
    image.line(x1, y1, x2, y2, col)


def text_with_border(image, x, y, s, col, bcol):
    for x_offset, y_offset in BORDER_OFFSETS:
        image.text(x + x_offset, y + y_offset, s, bcol)
    image.text(x, y, s, col)


def scale_image(image, scale):
    scaled_width = image.width * scale
    scaled_height = image.height * scale
    scaled_image = pyxel.Image(scaled_width, scaled_height)
    for y in range(scaled_height):
        for x in range(scaled_width):
            color = image.pget(x // scale, y // scale)
            scaled_image.pset(x, y, color)
    return scaled_image


def make_blt_figure():
    figure = pyxel.Image(pyxel.width, pyxel.height)
    figure.cls(BG_COLOR)
    image1 = pyxel.Image(32, 24)
    image1.blt(0, 0, 0, 0, 0, 32, 24)
    image1 = scale_image(image1, 3)
    image2 = pyxel.Image(32, 24)
    image2.blt(0, 0, 0, 0, 32, 32, 24)
    image2 = scale_image(image2, 3)

    col = 7
    bcol = 0

    def draw_w_and_h(x, y):
        pset_with_border(figure, x + 47, y + 23, col, bcol)
        line_with_border(figure, x + 47, y + 47, x + 70, y + 47, col, bcol)
        text_with_border(figure, x + 58, y + 44, "w", col, bcol)
        line_with_border(figure, x + 71, y + 23, x + 71, y + 46, col, bcol)
        text_with_border(figure, x + 70, y + 33, "h", col, bcol)

    x = 10
    y = 12
    figure.blt(x, y, image1, 0, 0, image1.width, image1.height)
    text_with_border(figure, x + 1, y - 7, "Screen", col, bcol)
    text_with_border(figure, x + 38, y + 16, "(x,y)", col, bcol)
    draw_w_and_h(x, y)

    x = 116
    figure.blt(x, y, image2, 0, 0, image2.width, image2.height)
    text_with_border(figure, x + 1, y - 7, "Image Bank", col, bcol)
    text_with_border(figure, x + 38, y + 16, "(u,v)", col, bcol)
    draw_w_and_h(x, y)

    return figure


def make_bltm_figure():
    figure = pyxel.Image(pyxel.width, pyxel.height)
    figure.cls(BG_COLOR)
    image1 = pyxel.Image(32, 24)
    image1.blt(0, 0, 0, 0, 64, 32, 24)
    image1 = scale_image(image1, 3)
    image2 = pyxel.Image(32, 24)
    image2.blt(0, 0, 0, 0, 96, 32, 24)
    image2 = scale_image(image2, 3)

    x = 10
    y = 12
    col = 7
    bcol = 0
    text_with_border(figure, x + 1, y - 7, "Tilemap", col, bcol)
    figure.blt(x, y, image1, 0, 0, image1.width, image1.height)

    col = 8
    bcol = 7
    text_with_border(figure, x + 3, y + 10, "(0,0) (0,0) (0,1) (0,2)", col, bcol)
    text_with_border(figure, x + 3, y + 34, "(3,2) (0,0) (0,0) (1,2)", col, bcol)
    text_with_border(figure, x + 3, y + 58, "(3,2) (2,2) (0,0) (0,2)", col, bcol)

    x = 116
    col = 7
    bcol = 0
    text_with_border(figure, x + 1, y - 7, "Image Bank (refimg)", col, bcol)
    figure.blt(x, y, image2, 0, 0, image2.width, image2.height)

    return figure


class App:
    def __init__(self):
        pyxel.init(223, 92, title="Offscreen Rendering")
        pyxel.load("assets/offscreen.pyxres")
        self.blt_figure = make_blt_figure()
        self.bltm_figure = make_bltm_figure()
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btnp(pyxel.KEY_Q):
            pyxel.quit()

    def draw(self):
        figure = (
            self.blt_figure if (pyxel.frame_count // 120) % 2 == 0 else self.bltm_figure
        )
        pyxel.blt(0, 0, figure, 0, 0, figure.width, figure.height)


App()



import pyxel

from .settings import PANEL_FOCUS_BORDER_COLOR, PANEL_FOCUS_COLOR
from .widgets import Widget


class TilemapViewer(Widget):
    """
    Variables:
        tilemap_no_var
        focus_x_var
        focus_y_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent, 157, 16, 66, 65)
        self._tilemap_image = pyxel.Image(64, 63)
        self.copy_var("tilemap_no_var", parent)
        self.copy_var("help_message_var", parent)

        # Initialize focus_x_var
        self.new_var("focus_x_var", 0)
        self.add_var_event_listener("focus_x_var", "set", self.__on_focus_x_set)

        # Initialize focus_y_var
        self.new_var("focus_y_var", 0)
        self.add_var_event_listener("focus_y_var", "set", self.__on_focus_y_set)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_focus(self, x, y):
        x = min(max((x - self.x - 1) // 2, 0), 31)
        y = min(max((y - self.y - 1) // 2, 0), 31)
        return x, y

    def __on_focus_x_set(self, value):
        return min(max(value, 0), 30)

    def __on_focus_y_set(self, value):
        return min(max(value, 0), 30)

    def __on_mouse_down(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self.focus_x_var, self.focus_y_var = self._screen_to_focus(x, y)

    def __on_mouse_drag(self, key, x, y, dx, dy):
        self.__on_mouse_down(key, x, y)

    def __on_mouse_hover(self, x, y):
        x, y = self._screen_to_focus(x, y)
        self.help_message_var = f"TARGET:CURSOR ({x * 8},{y * 8})"

    def __on_update(self):
        tilemap = pyxel.tilemap(self.tilemap_no_var)
        image = tilemap.image
        start_y = pyxel.frame_count % 8 * 8
        for y in range(start_y, start_y + 8):
            for x in range(64):
                tile = tilemap.pget(x * 4 + 1, y * 4 + 1)
                col = image.pget(tile[0] * 8 + 3, tile[1] * 8 + 3)
                self._tilemap_image.pset(x, y, col)

    def __on_draw(self):
        self.draw_panel(self.x, self.y, self.width, self.height)

        # Draw tilemap
        pyxel.blt(
            self.x + 1,
            self.y + 1,
            self._tilemap_image,
            0,
            0,
            self._tilemap_image.width,
            self._tilemap_image.height,
        )

        # Draw focus
        x = self.x + self.focus_x_var * 2 + 1
        y = self.y + self.focus_y_var * 2 + 1
        pyxel.clip(self.x + 1, self.y + 1, self.width - 2, self.height - 2)
        pyxel.rectb(x, y, 4, 4, PANEL_FOCUS_COLOR)
        pyxel.rectb(x - 1, y - 1, 6, 6, PANEL_FOCUS_BORDER_COLOR)
        pyxel.clip()
